######################
# Sommaire
######################
# utilisation : Eval buffer (C-c C-c) pour charger le paquet

# pour f endomorphisme nilpotent de type de Jordan fixé
# on calcule les dimensions des différentes familles de
# sous-espaces stables par f

# on rentre
# dimG = dimension du groupe
# dimV = dimension de la représentation
# orbites = liste des orbites nilpotentes, format [blocs, dim]

# on exécute
# Gr = liste_plausibles(dimG,dimV)
# donne la liste des grassmanniennes à tester
# puis
# teste(orbites, Gr)
# teste les orbites de la liste sur les grassmanniennes de la liste

#####
# Conventions pour l'ordre

# sauf mention explicite du contraire,
# j'écris les listes dans l'ordre de parcours, donc b_max en premier.
# exception notable : les fonctions foo_lisible écrites dans l'ordre de
# lecture normal d'un être humain

# la liste contenant les tailles des blocs de Jordan de f
# n'est pas ordonnée
# si f possède n blocs de tailles n_1,...,n_k c'est
# [n_1,...,n_k]
# exemple : f possède deux blocs de taille 3 et u bloc de taille 4
#           [3,3,4] ou [3,4,3], ou ...




######################
# List of all functions

# 1. Conversion between formats
# 2. Enumeration of types of stables, and dimension of families
# 3. Miscellaneous
######################



###############
# 1. Conversion between formats
# convert between various formats for encoding Jordan type
####

# jordan_lisible(blocs)
# arg blocs = liste contenant les tailles des blocs de Jordan de f
# sortie = liste [b_1,...,b_p] "type de Jordan"
#          b_i = nombre de blocs de taille i

# filtration(blocs)
# arg blocs = liste contenant les tailles des blocs de Jordan de f
# sortie = liste contenant les dim(K_i/K_{i-1}) "à l'envers",
#          dim(K_p/K_{p-1}) en premier (p indice de nilpotence)

# variante filtration_depuis_type(type)
# arg = type de Jordan [b_p,...,b_1]
# sortie = dim(K_i/K_{i-1}), i=p,...,1

# blocs(type)
# arg type = type de Jordan format [b_p,...,b_1]
#            (b_i nombre de blocs de taille i)
# sortie = liste contenant les tailles des blocs

# type(filtration)
# arg filtration= liste contenant les dim(K_i/K_{i-1}), i=p,...,1
# sortie = type de Jordan format [b_p,...,b_1]
#            (b_i nombre de blocs de taille i)

# noyaux_depuis_filtr(filtration)
# arg filtration= liste contenant les dim(K_i/K_{i-1}), i=p,...,1
# sortie = liste [k_p,...,k_1] des dimensions des noyaux des itérés 

# noyaux_depuis_type(type)
# même sortie que ci-dessus
# arg type = type de Jordan [b_p,...,b_1]



###############
# 2. Enumeration of types of stables, and dimension of families
# implements Proposition 6
####

# enumere_types(m,filtration)
# arg m = dimension des sous-espaces stables cherchés
# arg filtration= liste contenant les dim(K_i/K_{i-1}), i=p,...,1
# sortie = types de Jordan possibles pour F stable de dim=m
#          format [b_p,...,b_1]

# enumere_types_lisible(m,filtration)
# imprime tous les types obtenus par la fonction enumere_types
# au format de lecture humain [b_1,...,b_p]

# dim(type_F,filtr_f)
# arg type_F = type de Jordan de F [c_p,...,c_1]
# arg filtr_f = dims de la filtration pour f, k_i-k_{i-1}, i=p,...,1
# sortie = dimension de la famille de stables correspondante

# liste(m,blocs_f)
# arg m = dimension des sous-espaces stables
# arg blocs_f = liste des tailles des blocs de f
#               (ordre sans importance)
# sortie = liste des types de Jordan pour f_F avec la dimension des
#          familles correspondantes,
#          format [[c_p,...,c_1],dim_famille]

# liste_lisible(m,blocs_f)
# imprime la sortie de liste(m,blocs_f) au format de lecture humain
# [c_1,...,c_p], dim_famille



###############
# 3. Miscellaneous 
####

# somme(liste)
# somme les termes d'une liste d'entiers

# teste_somme(orbites,dimV)
# vérifie si les termes de la liste d'orbites
# ont des blocs de Jordan qui somment bien à dimV

######################################################################
# code
######

print("\nloading toolkit for nilpotent elements")

#ATTENTION
# j'écris les listes dans l'ordre de parcours, donc b_max en premier

def max_liste(l):
    m=0
    for a in l:
        m=max(m,a)
    return m

def jordan_lisible(blocs):
    l=[0]*max_liste(blocs)
    for b in blocs:
        l[b-1]=l[b-1]+1
    return l

# pour la sortie :
# dim (K_p/K_{p-1}) en premier, p indice de nilpotence
def filtration(blocs):
    l=[0]*max_liste(blocs)
    for b in blocs:
        for i in range(b):
            l[i]=l[i]+1
    return l[::-1]

# variante : calcul direct depuis le type
def filtration_depuis_type(type):
    filtre=type[-len(type):]
    #copie a à partir de filtration[1]
    for i in range(1,len(filtre)):
        filtre[i]=filtre[i-1]+filtre[i]
        i=i+1
    return filtre

def blocs(type):
    l=[]
    t=len(type)
    for b in type:
        l=l+[t]*b
        t=t-1
    return l

def type(filtration):
    l=filtration[-len(filtration):]
    for i in range(len(l)-1,0,-1):
        l[i]=l[i]-l[i-1]
    return l

def noyaux_depuis_filtr(filtration):
    long=len(filtration)
    dims=[filtration[long-1]]
    for i in range(long-1):
        dims=[filtration[long-2-i]+dims[0]]+dims
    return dims

def noyaux_depuis_type(type):
    long=len(type)
    dims=[]
    for i in range(1,long+1):
        ki=0
        for j in range(1,long+1):
            ki=ki+min(i,j)*type[long-j]
        dims=[ki]+dims
    return dims

def enumere_types(m,filtration):
    if len(filtration)==1:
        if m<=filtration[0]:
            return [[m]]
        if m>filtration[0]:
            return []
    catalogo=[]
    for b in range(filtration[0]+1):
        n=m-b*len(filtration)
        if n>=0:
            x=filtration[-len(filtration)+1:]
            #copie a à partir de filtration[1]
            for i in range(len(x)):
                x[i]=x[i]-b
            retour=enumere_types(n,x)
            for i in range(len(retour)):
                retour[i]=[b]+retour[i]
            catalogo=catalogo+retour
    return catalogo

def enumere_types_lisible(m,filtration):
    liste=enumere_types(m,filtration)
    for type in liste:
        print (type[::-1])

def dim(type_F,filtr_f):
    noyaux_F = noyaux_depuis_type(type_F)
    noyaux_f = noyaux_depuis_filtr(filtr_f)
    dim=0
    for i in range(len(type_F)):
        dim = dim + type_F[i]*(noyaux_f[i]-noyaux_F[i])
    return dim

def liste(m,blocs_f):
    filtr=filtration(blocs_f)
    l=enumere_types(m,filtr)
    for i in range(len(l)):
        l[i]=[l[i]]+[dim(l[i],filtr)]
    return l

def liste_lisible(m,blocs_f):
    sortie = liste(m,blocs_f)
    for x in sortie:
        print(x[0][::-1],x[1])
        # imprime dans le sens normal de lecture
        # b1 en premier

def somme(liste):
    s=0
    for x in liste:
        s=s+x
    return s

def teste_somme(orbites,dimV):
    test = True
    for o in orbites:
        if somme(o[0]) != dimV:
            test = False
            print("erreur orbite",o[0])
    if test:
        print("tutt'ok")

        
# implémentation tests automatiques

# liste les cas pour lesquels il est plausible que
# G = dim du groupe
# dimV = dim de la représentation
def liste_plausibles(dimG,dimV):
    l=[]
    for i in range (dimV+1):
        dim = i*(dimV-i)
        if dim >= dimG:
            l = l + [[i,dim]]
    return l

def print_plausibles(l):
    print("dimensions for which it is plausible that the general F has trivial stabilizer\n")
    for x in l:
        print("dim(F) =", x[0], ", corresp. Grass. has dim =", x[1])


# orbites = liste des orbites nilpotentes
# Gr = liste des [m,dim] où m = dim des sev,
#                           dim = dim de la Grassmannienne
def teste_nilp(orbites, Gr):
    for gr in Gr:
        test = True
        print("m=",gr[0])
        for orb in orbites:
            l=liste(gr[0],orb[0])
            for x in l:
                if x[1]+orb[1]>gr[1]:
                    test = False
                    print("depasse pour orbite",jordan_lisible(orb[0]),orb[1])
                    print("et type de stable", x[0][::-1],x[1])
        if test:
            print("OK")

def teste_nilp_bavard(orbites, Gr):
    test = True
    for gr in Gr:
        mx = 0
        print("\n***************\n")
        print("dim(F) =",gr[0])
        print("dim(Gr) = ", gr[1])
        for orb in orbites:
            print("\n", "orbit : blocks ",orb[0]," and orb.dim=",orb[1],sep='')
            l=liste(gr[0],orb[0])
            for x in l:
                print("stable subspaces with blocks", blocs(x[0]), "move in dimension", x[1])
                mx = max(mx, x[1])
            print("largest family has dim",mx)
            if mx<=gr[1]-orb[1]:
                print("indeed <= dim(Gr)-dim(orb)")
            else:
                test = False
                print("whoops > dim(Gr)-dim(orb)")
    if test:
        print("\n\nOK in all cases, viz.,",Gr[0][0],"<= dim(F) <=",Gr[len(Gr)-1][0])



##################
# definition of complete procedure
# (cosmetic presentation)

def nilp_complete (g, G, dimG, V, dimV, orbites, synthetic):
    print("\n"+60*"*")
    print("genus", g)
    print("unipotent case")
    print("")
    print ("group "+G+" : ", "dim =", dimG)
    print ("representation "+V+" : ", "dim =", dimV)
    #
    print("\n"+40*"*")
    print("\nlist of non-trivial nilpotent orbits")
    for x in orbites:
        print("orb.dim",x[1],", blocks",x[0])
    #
    Gr = liste_plausibles(dimG,dimV)
    print("\n"+40*"*")
    print_plausibles(Gr)
    # now let's go
    print("\n"+40*"*")
    print("\nnow let's enumerate all possibilities and see if they pass the test STAR")
    #
    if synthetic:
        teste_nilp(orbites, Gr)
    else:
        teste_nilp_bavard(orbites, Gr)
    print("\n"+60*"*")

def test():
    print("nilp")

print("OK toolkit loaded")
